#pragma once

#include "table.h"

namespace cpphibernate {
namespace mariadb {

    /**
     * @brief Table for polymorphic data types.
     */
    template<typename T_schema, typename T_table>
    struct table_polymorphic_t
        : public table_t
    {
    public:
        using dataset_type = typename T_table::dataset_type;

    public:
        /**
         * @brief Value constructor. Creates a mariadb table from the cpphibernate table.
         *
         * @param[in] p_owner   Owner of the table.
         * @param[in] p_schema  Cpphibernate schema the table belongs to.
         * @param[in] p_table   Cpphibernate table to create mariadb table for.
         */
        template<
            typename T_base_dataset,
            typename T_derived_datasets>
        inline table_polymorphic_t(
            const schema_t& p_owner,
            const T_schema& p_schema,
            const T_table&  p_table,
            const T_base_dataset&&,
            const T_derived_datasets&&);

    public:
        /**
         * @brief Execute a create/update operation on the current table.
         *
         * For a polymorphic type this will check the derived tables before executing the actual opertion.
         * If the dataset matches one of the dericed tables, the operation is executed on this table.
         * This operation also updates the base table if the dataset has one.
         *
         * @param[in] context   Context that stores the needed data for the operation.
         *
         * @return              Returns the key of the created/updated dataset in it's string representation.
         */
        std::string create_update(const create_update_context& context) const override;

        /**
         * @brief Execute a destroy operation on the current table.
         *
         * For a polymorphic type this will check the derived tables before executing the actual operation.
         * If the dataset matches one of the dericed tables, the operation is executed on this table.
         * This operation also updates the base table if the dataset has one.
         *
         * @param[in] context   Context that stores the needed data for the operation.
         */
        void destroy(const destroy_context& context) const override;

    public:
        /**
         * @brief Emplace new dataset of the type the table represents.
         *
         * @param[in] context   Context to emplace new dataset in.
         */
        void emplace(const read_context& context) const override;

    private:
        /**
         * @brief Execute the predicate for each derived type.
         *
         * @param[in] dataset       Dataset to try to cast to derived type.
         * @param[in] include_self  True: Also check the dataset type. False: Only check derived types.
         * @param[in] pred          Predicate to execute if dataset could be cast to a derived type.
         */
        template<typename T_dataset, typename T_pred, typename T_include_self>
        constexpr void for_each_derived(T_dataset& dataset, const T_include_self& include_self, const T_pred& pred) const;
    };

} }
