#pragma once

#include "table_polymorphic.h"

#include "../../context/create_update_context.inl"
#include "../../context/destroy_context.inl"

namespace cpphibernate {
namespace mariadb {

    namespace __impl
    {

        /* make_dataset_id_vector */

        struct make_dataset_id_vector_impl
        {
            template<typename T_datasets, size_t... I>
            static constexpr decltype(auto) helper(const T_datasets& datasets, std::index_sequence<I...>)
            {
                return std::vector<size_t>({
                    get_type_id(datasets[hana::size_c<I>])...
                });
            }

            template<typename T_datasets>
            constexpr decltype(auto) operator()(const T_datasets& datasets) const
            {
                using size = mp::decay_t<decltype(hana::size(datasets))>;
                return helper(datasets, std::make_index_sequence<size::value> { });
            }
        };

        static constexpr decltype(auto) make_dataset_id_vector = make_dataset_id_vector_impl { };

        /* filter_own_type_impl */

        template<typename T_dataset, typename T_include_self>
        struct filter_self_type_impl
        {
            template<typename T_type>
            inline decltype(auto) operator() (T_type&&) const
            {
                return hana::and_(
                    hana::not_(hana::trait<std::is_abstract>(T_type{ })),
                    hana::or_(
                        T_type{ } != hana::type_c<T_dataset>,
                        T_include_self { }));
            }
        };

        template<typename T_dataset, typename T_include_self>
        static constexpr decltype(auto) filter_self_type = filter_self_type_impl<T_dataset, T_include_self> { };

    }

    /* table_polymorphic_t */

    template<
        typename T_schema,
        typename T_table>
    template<
        typename T_base_dataset,
        typename T_derived_datasets>
    table_polymorphic_t<T_schema, T_table>
        ::table_polymorphic_t(
            const schema_t& p_owner,
            const T_schema& p_schema,
            const T_table&  p_table,
            const T_base_dataset&&,
            const T_derived_datasets&&)
            : table_t(
                p_owner,
                p_schema,
                p_table)
    {
        this->base_dataset_id = hana::if_(
            hana::equal(T_base_dataset { }, hana::type_c<void>),
            static_cast<size_t>(0),
            get_type_id(T_base_dataset { }));
        this->derived_dataset_ids = __impl::make_dataset_id_vector(T_derived_datasets { });
    }

    template<
        typename T_schema,
        typename T_table>
    std::string table_polymorphic_t<T_schema, T_table>
        ::create_update(const create_update_context& context) const
    {
        bool  done    = false;
        auto& dataset = context.get<dataset_type>();
        for_each_derived(dataset, hana::false_c, [&](auto& derived_dataset){
            if (!done)
            {
                using derived_dataset_type  = mp::decay_t<decltype(derived_dataset)>;
                auto  derived_dataset_id    = get_type_id(hana::type_c<derived_dataset_type>);
                auto  derived_table         = this->get_derived_by_dataset_id(derived_dataset_id);
                if (!derived_table)
                {
                    throw exception(static_cast<std::ostringstream&>(std::ostringstream { }
                        << "unable to find derived table info for dataset '"
                        << cppcore::type_helper<derived_dataset_type>::name() << "'!").str());
                }
                derived_table->create_update(change_context(context, derived_dataset));
                done = true;
            }
        });

        return done
            ? *this->primary_key_field->get(context)
            : this->create_update_exec(context);
    }

    template<
        typename T_schema,
        typename T_table>
    void table_polymorphic_t<T_schema, T_table>
        ::destroy(const destroy_context& context) const
    {
        bool  done    = false;
        auto& dataset = context.get<dataset_type>();
        for_each_derived(dataset, hana::false_c, [&](auto& derived_dataset){
            if (!done)
            {
                using derived_dataset_type  = mp::decay_t<decltype(derived_dataset)>;
                auto  derived_dataset_id    = get_type_id(hana::type_c<derived_dataset_type>);
                auto  derived_table         = this->get_derived_by_dataset_id(derived_dataset_id);
                if (!derived_table)
                {
                    throw exception(static_cast<std::ostringstream&>(std::ostringstream { }
                        << "unable to find derived table info for dataset '"
                        << cppcore::type_helper<derived_dataset_type>::name() << "'!").str());
                }
                auto new_context = change_context(context, derived_dataset);
                derived_table->destroy(new_context);
                done = true;
            }
        });

        if (!done)
        {
            this->destroy_exec(context);
        }
    }

    template<
        typename T_schema,
        typename T_table>
    void table_polymorphic_t<T_schema, T_table>
        ::emplace(const read_context& context) const
    {
        hana::eval_if(
            std::is_abstract_v<dataset_type>,
            [](){
                throw exception(std::string("can not create dataset of abstract type: ")
                    + cppcore::type_helper<dataset_type>::name());
            },
            [&context, this](auto _){
                _(context).template emplace<dataset_type>(this);
            });
    }

    template<
        typename T_schema,
        typename T_table>
    template<
        typename T_dataset,
        typename T_pred,
        typename T_include_self>
    constexpr void table_polymorphic_t<T_schema, T_table>
        ::for_each_derived(T_dataset& dataset, const T_include_self& include_self, const T_pred& pred) const
    {
        auto derived_types = decltype(hana::filter(
            schema::get_all_derived_types(std::declval<T_schema>(), hana::type_c<dataset_type>),
            __impl::filter_self_type<mp::decay_t<T_dataset>, mp::decay_t<T_include_self>>)) { };

        hana::for_each(derived_types, [&](auto& type){
            using derived_type = decay_unwrap_t<decltype(type)>;
            auto* derived      = dynamic_cast<derived_type*>(&dataset);
            if (derived)
                pred(*derived);
        });
    }

} }
