#pragma once

#include "data_context.h"
#include "../impl/filter.h"

namespace cpphibernate {
namespace mariadb {

    /**
     * @brief Context that is used for read operations.
     */
    struct read_context
        : public data_context
    {
    protected:
        const size_t        real_dataset_id;    //!< Unique id of thre real dataset the context was created with.

    public:
        const filter_t&     filter;             //!< Filter that is used for read operation.
        bool                is_dynamic;         //!< The dataset managed by this context is dynamic/polymorhp
        std::string         where;              //!< Where statement to use for selection
        std::string         limit;              //!< Limit statement to use for selection
        std::string         order_by;           //!< Order statement to use for selection

    public:
        /**
         * @brief Destructor.
         */
        virtual ~read_context() = default;

        /**
         * @brief Emplace a new value in the dataset of the read context.
         *
         * @tparam T_dataset    Type of the dataset to emplace.
         *
         * @param[in] table     Table of the dataset to emplace.
         *                      If the table is unknown, the table is searched within the schema.
         *
         * @return              Reference to the emplaced dataset.
         */
        template<typename T_dataset>
        inline T_dataset& emplace(const table_t * table = nullptr) const;

        /**
         * @brief Emplace a new value in the dataset of the read context.
         */
        inline void emplace() const;

        /**
         * @brief Finish the emplacement of values in the dataset.
         */
        inline void finish() const;

    private:
        /**
         * @brief Actual emplace method. This method should create a new value entry in the dataset of the read context.
         *
         * @param[in] data          Actual data to emplace. If nullptr is passed a default value is created.
         * @param[in] dataset_id    Unique id of the dataset to emplace (must match the object pointed to by data).
         *
         * @return                  Pointer to the dataset that was emplaced.
         */
        virtual void * emplace_intern(void * data, size_t dataset_id) const = 0;

        /**
         * @brief Actual finish method. Should check how many element has been emplaced in the dataset.
         */
        virtual void finish_intern () const = 0;

    protected:
        /**
         * @brief Constructor. Creates a create context (no filters are passed).
         *
         * @param[in] p_schema              Mariadb driver schema to use for the operation.
         * @param[in] p_connection          Mariadb connection to execute queries with.
         * @param[in] p_filter              Filter to use for the read operation.
         * @param[in] p_real_dataset_id     Unique id the dataset that is stored in the context.
         */
        inline read_context(
            const schema_t&             p_schema,
            ::cppmariadb::connection&   p_connection,
            const filter_t&             p_filter,
            size_t                      p_real_dataset_id);
    };

    using read_context_ptr_u = std::unique_ptr<read_context>;

    namespace __impl
    {

        /**
         * @brief Helper class to create read contexts.
         */
        template<typename X, typename = void>
        struct read_context_builder;

    }

    /**
     * @brief Predicate to create read contexts.
     */
    constexpr decltype(auto) make_read_context = mp::generic_predicate<__impl::read_context_builder> { };

} }
