#pragma once

#include "impl/driver_impl.h"

namespace cpphibernate {
namespace mariadb {

    /**
     * @brief Mariadb driver.
     *
     * The normal cpphibernate context will be inherited from this class. So all methods
     * defined here will be part of the context class.
     */
    struct driver_t
    {
    public:
        friend struct driver_impl_t;

    private:
        driver_impl_t               _impl;          //!< Driver implementation.
        ::cppmariadb::connection *  _connection;    //!< Mariadb connection to use for queries.
        filter_t                    _filter;        //!< Filter to use for read and update operations.

    public:
        /**
         * @brief Constructor. Initializes the driver with the passed schema.
         *
         * @param[in] p_schema  Schema to use with this driver.
         */
        template<typename T_schema>
        inline driver_t(T_schema&& p_schema);

        /**
         * @brief Constructor. Initializes the driver with the passed schema.
         *
         * @param[in] p_schema      Schema to use with this driver.
         * @param[in] p_connection  Connection to the mariadb server to use for queries.
         */
        template<typename T_schema>
        inline driver_t(T_schema&& p_schema, ::cppmariadb::connection& p_connection);

        /**
         * @brief Print the schema of the driver to the passed stream.
         */
        inline std::ostream& print(std::ostream& os) const;

        /**
         * @brief Get the connection currently assigned to the driver.
         */
        inline ::cppmariadb::connection * connection() const;

        /**
         * @brief Set the connection to use for queries.
         *        This will invalidate all cached queries and strings.
         */
        inline void connection(::cppmariadb::connection * p_connection);

        /**
         * @brief Set the include filter to use for update and read operations.
         *
         * @param[in] args  Fileds and tables to use as include filter.
         */
        template<typename... T_args>
        inline void set_filter_inclusive(T_args&&... args);

        /**
         * @brief Set the exclude filter to use for update and read operations.
         *
         * @param[in] args  Fileds and tables to use as exclude filter.
         */
        template<typename... T_args>
        inline void set_filter_exclusive(T_args&&... args);

        /**
         * @brief Clear the filter.
         */
        inline void clear_filter();

    protected:
        /**
         * @brief Get the imlementation object of the driver.
         */
        inline auto& impl() const;
    };

} }
