#pragma once

#include <cpphibernate/config.h>

namespace cpphibernate {
namespace schema {

    namespace __impl
    {

        /**
         * @brief Tag class to mark schema objects.
         */
        struct tag_schema
            { };

        /**
         * @brief Schema type with all tables and fields for a database layout.
         *
         * @tparam T_name   Name of the schema.
         * @tparam T_tables List of all tables that belong to this schema.
         */
        template<typename T_name, typename T_tables>
        struct schema_t
            : public tag_schema
            , public tag_equality_comparable<tag_schema>
        {
            using name_type     = T_name;
            using tables_type   = T_tables;

            name_type   name;   //!< Name of the schema.
            tables_type tables; //!< List of all tables that belongs to the schema.

            /**
             * @brief Constructor.
             *
             * @param[in] p_name    Name of the schema.
             * @param[in] p_tables  List of all tables that belongs to the schema.
             */
            constexpr schema_t(
                name_type&&    p_name,
                tables_type&&  p_tables);

            /**
             * @brief Move constructor.
             */
            constexpr schema_t(schema_t&&) = default;

            /**
             * @brief Copy constructor.
             */
            constexpr schema_t(const schema_t&) = delete;

            /**
             * @brief Move assignment constructor.
             */
            constexpr schema_t& operator = (schema_t&&) = default;

            /**
             * @brief Copy assignment constructor.
             */
            constexpr schema_t& operator = (const schema_t&) = delete;

            /**
             * @brief Print the whole schema to the passed stream.
             */
            inline void print(std::ostream& os) const;
        };

        /**
         * @brief Helper type to build schema objects.
         */
        template<typename X, typename = void>
        struct schema_builder;

        /**
         * @brief Helper type to get the base type of the passed dataset type.
         */
        template<typename X, typename = void>
        struct get_base_type_builder;

        /**
         * @brief Helper class to get a list of the directly derived types of the passed dataset type.
         */
        template<typename X, typename = void>
        struct get_derived_types_builder;

        /**
         * @brief Helper class to get a list of all derived types (including the passed dataset type).
         */
        template<typename X, typename = void>
        struct get_all_derived_types_builder;

    }

    /**
     * @brief Evaluates to true_t if the passed type is a schema type.
     */
    template<typename T>
    struct is_schema;

    /**
     * @brief Is true if the passed type is an attributes type, false otherwise.
     */
    template<typename T>
    constexpr bool is_schema_v = is_schema<T>::value;

    /**
     * @brief Predicate to create an schema object from the passed attributes.
     */
    constexpr decltype(auto) make_schema = mp::generic_predicate<__impl::schema_builder> { };

    /**
     * @brief Predicate to get the base type of the passed dataset type.
     */
    constexpr decltype(auto) get_base_type = mp::generic_predicate<__impl::get_base_type_builder> { };

    /**
     * @brief Predicate to get a list of the directly derived types of the passed dataset type.
     */
    constexpr decltype(auto) get_derived_types = mp::generic_predicate<__impl::get_derived_types_builder> { };

    /**
     * @brief Predicate to get a list of all derived types (including the passed dataset type).
     */
    constexpr decltype(auto) get_all_derived_types = mp::generic_predicate<__impl::get_all_derived_types_builder> { };

} }

#include "schema.inl"
