#pragma once

#include <cpphibernate/misc/type_helper.h>

#include "fields.h"
#include "table.h"

namespace cpphibernate {
namespace schema {

    namespace __impl
    {

        /* table_t */

        template<typename T_name, typename T_wrapped_dataset, typename T_table_id, typename T_fields>
        constexpr table_t<T_name, T_wrapped_dataset, T_table_id, T_fields>
            ::table_t(
                name_type              p_name,
                wrapped_dataset_type   p_wrapped_dataset,
                table_id_type          p_table_id,
                fields_type            p_fields)
                : name              (std::forward<name_type>           (p_name))
                , wrapped_dataset   (std::forward<wrapped_dataset_type>(p_wrapped_dataset))
                , table_id          (std::forward<table_id_type>       (p_table_id))
                , fields            (std::forward<fields_type>         (p_fields))
                { }

        template<typename T_name, typename T_wrapped_dataset, typename T_table_id, typename T_fields>
        void table_t<T_name, T_wrapped_dataset, T_table_id, T_fields>
            ::print(std::ostream& os) const
        {
            using namespace ::cppcore;
            size_t index = 0;
            os  << indent << '{'
                << incindent
                    << indent << "\"name\": \"" << name << "\","
                    << indent << "\"wrapped_dataset\": \"" << type_helper<dataset_type>::name() << "\""
                    << indent << "\"table_id\": \"" << hana::value(table_id) << "\""
                    << indent << "\"fields\": "
                    << indent << '['
                    << incindent;
                    hana::for_each(fields, [&](auto& table){
                        if (index++ > 0) os << ",";
                        table.print(os);
                    });
            os      << decindent
                    << indent << ']'
                << decindent
                << indent << '}';
        }

        /* table_builder */

        template<typename X, typename>
        struct table_builder
        {
            template<typename... T_args>
            static constexpr decltype(auto) apply(T_args&&...)
                { static_assert(sizeof...(T_args) == -1, "Invalid parameters for make_table(...)!"); }
        };

        template<
            typename T_name,
            typename T_wrapped_dataset,
            typename T_table_id,
            typename T_fields>
        struct table_builder<
            mp::list<T_name, T_wrapped_dataset, T_table_id, T_fields>,
            mp::enable_if_t<
                    mp::is_valid_v<unwrap_t<mp::decay_t<T_wrapped_dataset>>>
                &&  is_fields_v<mp::decay_t<T_fields>>>>
        {
            using name_type             = T_name;
            using wrapped_dataset_type  = T_wrapped_dataset;
            using table_id_type         = T_table_id;
            using fields_type           = T_fields;
            using table_type            = table_t<name_type, wrapped_dataset_type, table_id_type, fields_type>;

            static constexpr decltype(auto) apply(
                name_type&&             name,
                wrapped_dataset_type&&  wrapped_dataset,
                table_id_type&&         table_id,
                fields_type&&           fields)
            {
                return table_type(
                    std::forward<name_type>             (name),
                    std::forward<wrapped_dataset_type>  (wrapped_dataset),
                    std::forward<table_id_type>         (table_id),
                    std::forward<fields_type>           (fields));
            }
        };

        /* get_wrapped_dataset */

        template<typename X, typename>
        struct get_wrapped_dataset_builder
        {
            template<typename... T_args>
            static constexpr decltype(auto) apply(T_args&&...)
                { static_assert(sizeof...(T_args) == -1, "Invalid parameters for get_wrapped_dataset(...)!"); }
        };

        template<typename T_table>
        struct get_wrapped_dataset_builder<
            mp::list<T_table>,
            mp::enable_if_t<is_table_v<mp::decay_t<T_table>>>>
        {
            static constexpr decltype(auto) apply(T_table&& table)
                { return std::forward<T_table>(table).wrapped_dataset; }
        };

        /* get_primary_key_field_builder */

        template<typename X, typename>
        struct get_primary_key_field_builder
        {
            template <typename... T_args>
            static constexpr decltype(auto) apply(T_args&&... args)
                { static_assert(sizeof...(args) == -1, "Invalid parameters for get_primary_key_field(...)!"); }
        };

        template<typename T_table>
        struct get_primary_key_field_builder<
            mp::list<T_table>,
            mp::enable_if_t<
                is_table_v<mp::decay_t<T_table>>>>
        {
            template<size_t I>
            using is_primary_key_field = decltype(hana::contains(
                std::declval<T_table>().fields[hana::size_c<I>].attributes,
                attribute::primary_key));

            template<size_t I, size_t N, typename = void>
            struct helper;

            template<size_t N>
            struct helper<N, N, void>
                { static_assert(N != N, "Unable to find primary key field for table!"); };

            template<size_t I, size_t N>
            struct helper<I, N, mp::enable_if_t<
                !is_primary_key_field<I>::value>>
            {
                static constexpr decltype(auto) apply(T_table&& table)
                    { return helper<I+1, N>::apply(std::forward<T_table>(table)); }
            };

            template<size_t I, size_t N>
            struct helper<I, N, mp::enable_if_t<
                is_primary_key_field<I>::value>>
            {
                static constexpr decltype(auto) apply(T_table&& table)
                    { return std::forward<T_table>(table).fields[hana::size_c<I>]; }
            };

            static constexpr decltype(auto) apply(T_table&& table)
            {
                using count = mp::decay_t<decltype(hana::size(std::declval<T_table>().fields))>;
                return helper<0, count::value>::apply(
                    std::forward<T_table>(table));
            }
        };

    }

    /* is_table */

    template<typename T>
    struct is_table
        : public mp::is_specialization_of<T, __impl::table_t>
        { };

} }
