#include <list>
#include <cpphibernate/driver/mariadb/classes/tables/table.h>
#include <cpphibernate/driver/mariadb/context/base_context.inl>
#include <cpphibernate/driver/mariadb/context/destroy_context.inl>

using namespace ::cpphibernate;
using namespace ::cpphibernate::mariadb;

struct delete_query_builder_t
{
    const table_t&          this_table;

    size_t                  alias_id { 0 };
    std::ostringstream      os;
    std::list<std::string>  names;

    delete_query_builder_t(
        const table_t& p_this_table)
        : this_table(p_this_table)
        { }

    inline std::string make_alias()
        { return std::string("T") + std::to_string(alias_id++); }

    inline void add_table(const table_t& table, const std::string& alias, bool add_base, bool add_derived)
    {
        auto has_alias  = !alias.empty();
        auto real_alias = has_alias ? alias : table.name;

        if (table.base_table && add_base)
        {
            assert(table.base_table->primary_key_field);

            auto& base_table      = *table.base_table;
            auto& base_key        = *base_table.primary_key_field;
            auto  base_alias      = has_alias ? make_alias() : std::string();
            auto  real_base_alias = has_alias ? base_alias : base_table.name;

            os  <<  " LEFT JOIN `"
                <<  base_table.name;
            if (has_alias)
            {
                os  <<  "` AS `"
                    <<  base_alias;
            }
            os  <<  "` ON `"
                <<  real_alias
                <<  "`.`"
                <<  base_key.name
                <<  "`=`"
                <<  real_base_alias
                <<  "`.`"
                <<  base_key.name
                <<  "`";

            add_table(base_table, base_alias, true, false);
        }

        names.emplace_back(real_alias);

        /* foreign table one */
        for (auto& ptr : table.foreign_table_one_fields)
        {
            assert(ptr);
            assert(ptr->table.primary_key_field);
            assert(ptr->referenced_table);
            assert(ptr->referenced_table->primary_key_field);

            auto& field     = *ptr;
            auto& own_key   = *table.primary_key_field;
            auto& ref_table = *field.referenced_table;
            auto& ref_key   = *ref_table.primary_key_field;
            auto  new_alias = make_alias();

            if (ref_table.is_used_in_container)
            {
                os  <<  " LEFT JOIN `"
                    <<  ref_table.name
                    <<  "` AS `"
                    <<  new_alias
                    <<  "` ON `"
                    <<  real_alias
                    <<  "`.`"
                    <<  own_key.name
                    <<  "`=`"
                    <<  new_alias
                    <<  "`.`"
                    <<  table.name
                    <<  "_id_"
                    <<  field.name
                    <<  "`";
            }
            else
            {
                os  <<  " LEFT JOIN `"
                    <<  ref_table.name
                    <<  "` AS `"
                    <<  new_alias
                    <<  "` ON `"
                    <<  real_alias
                    <<  "`.`"
                    <<  ref_key.table.name
                    <<  "_id_"
                    <<  field.name
                    <<  "`=`"
                    <<  new_alias
                    <<  "`.`"
                    <<  ref_key.name
                    <<  "`";
            }

            add_table(ref_table, new_alias, true, true);
        }

        /* derived tables */
        if (add_derived)
        {
            for (auto& ptr : table.derived_tables)
            {
                assert(ptr);
                assert(ptr->primary_key_field);

                auto& derived_table      = *ptr;
                auto& primary_key        = *table.primary_key_field;
                auto  derived_alias      = has_alias ? make_alias() : std::string();
                auto  real_derived_alias = has_alias ? derived_alias : derived_table.name;

                os  <<  " LEFT JOIN `"
                    <<  derived_table.name;
                if (has_alias)
                {
                    os  <<  "` AS `"
                        <<  derived_alias;
                }
                os  <<  "` ON `"
                    <<  real_alias
                    <<  "`.`"
                    <<  primary_key.name
                    <<  "`=`"
                    <<  real_derived_alias
                    <<  "`.`"
                    <<  primary_key.name
                    <<  "`";

                add_table(derived_table, derived_alias, false, true);
            }
        }
    }

    inline std::string operator()(const std::string* where)
    {
        os  <<  " FROM `"
            <<  this_table.name
            <<  "`";
        add_table(this_table, "", true, true);
        if (where)
        {
            os << " " << *where;
        }
        else
        {
            os << " ?where!";
        }

        std::ostringstream ss;
        ss  <<  "DELETE ";
        bool first = true;
        for (auto& name : names)
        {
            if (first)
                first = false;
            else
                ss << ", ";
            ss  <<  "`"
                <<  name
                <<  "`";
        }
        ss << os.str();

        return ss.str();
    }
};

static std::string build_foreign_many_delete_query(const table_t& table);

void table_t::destroy(const destroy_context& context) const
    { return destroy_exec(context); }

void table_t::cleanup(const base_context& context, table_set& processed, bool check_derived, bool check_base) const
{
    /* check and add the table to the set of processed tables */
    if (processed.count(this))
        return;
    processed.emplace(this);

    /* execute the cleanup query */
    foreign_many_delete_exec(context);

    /* cleanup all foreign fields */
    for (auto ptr : foreign_table_fields)
    {
        assert(ptr);
        assert(ptr->referenced_table);
        auto& ref_table = *ptr->referenced_table;
        if (ref_table.is_used_in_container)
            ref_table.cleanup(context, processed, true, true);
    }

    /* cleanup the base table */
    if (check_base && base_table)
    {
        base_table->cleanup(context, processed, false, true);
    }

    /* cleanup the dereived tables */
    if (check_derived)
    {
        for (auto ptr : derived_tables)
        {
            assert(ptr);
            ptr->cleanup(context, processed, true, false);
        }
    }
}

void table_t::destroy_exec(const destroy_context& context) const
{
    assert(primary_key_field);

    auto& connection = context.connection;
    auto& statement  = get_statement_delete();

    statement.set(0, context.where);
    cpphibernate_log_debug("execute DELETE query: " << std::endl << statement.query(connection) << std::endl);
    connection.execute(statement);

    table_set processed;
    for (auto& ptr : foreign_table_fields)
    {
        assert(ptr);
        assert(ptr->referenced_table);
        auto& ref_table = *ptr->referenced_table;
        if (ref_table.is_used_in_container)
            ref_table.cleanup(context, processed, true, true);
    }
}

void table_t::foreign_many_delete_exec(const base_context& context) const
{
    if (foreign_key_fields.empty())
        return;
    auto& connection = context.connection;
    auto& statement  = get_statement_foreign_many_delete();
    cpphibernate_log_debug("execute DELETE old foreign many query: " << std::endl << statement.query(connection) << std::endl);
    connection.execute(statement);
}

std::string table_t::build_delete_query(const std::string* where) const
    { return delete_query_builder_t(*this)(where); }

::cppmariadb::statement& table_t::get_statement_foreign_many_delete() const
{
    if (!_statement_foreign_many_delete)
    {
        auto query = build_foreign_many_delete_query(*this);
        _statement_foreign_many_delete.reset(new ::cppmariadb::statement(std::move(query)));
    }
    return *_statement_foreign_many_delete;
}

::cppmariadb::statement& table_t::get_statement_delete() const
{
    if (!_statement_delete)
    {
        auto query = delete_query_builder_t(*this)(nullptr);
        _statement_delete.reset(new ::cppmariadb::statement(std::move(query)));
    }
    return *_statement_delete;
}

std::string build_foreign_many_delete_query(const table_t& table)
{
    std::ostringstream os;
    auto first = true;
    os << "WHERE";
    for (auto ptr : table.foreign_key_fields)
    {
        assert(ptr);
        auto& field = *ptr;
        if (first)
            first = false;
        else
            os << " AND";
        os  <<  " (`"
            <<  field.table.name
            <<  "_id_"
            <<  field.name
            <<  "` IS NULL)";
    }
    std::string where = os.str();

    return delete_query_builder_t(table)(&where);
}
